// ColorBalance - Avisynth plugin to mimic the Gimp's Color Balance tool
// Initial prototype by Gavino (http://forum.doom9.org/member.php?u=141297)

/* GIMP - The GNU Image Manipulation Program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 */

// Avisynth v2.5.  Copyright 2002 Ben Rudiak-Gould et al.
// http://www.avisynth.org

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .
//
// Linking Avisynth statically or dynamically with other modules is making a
// combined work based on Avisynth.  Thus, the terms and conditions of the GNU
// General Public License cover the whole combination.
//
// As a special exception, the copyright holders of Avisynth give you
// permission to link Avisynth with independent modules that communicate with
// Avisynth solely through the interfaces defined in avisynth.h, regardless of the license
// terms of these independent modules, and to copy and distribute the
// resulting combined work under terms of your choice, provided that
// every copy of the combined work is accompanied by a complete copy of
// the source code of Avisynth (the version of Avisynth used to produce the
// combined work), being distributed under the terms of the GNU General
// Public License plus this exception.  An independent module is a module
// which is not derived from or based on Avisynth, such as 3rd-party filters,
// import and export plugins, or graphical user interfaces.


// Change History:

// v0.1 (Gavino, 26th August 2008):
// - initial code based on Gimp
// v0.2 (Mikeytown2, 27th August 2008):
// - Added preserve luminosity from Gimp & got rid of all warnings in MS VC++ 2005
// v0.21 (Mikeytown2, 27th August 2008):
// - Code Cleanup (tidy & logic(min,max)); Got rid of all if's from loops; Gimp support functions not in the ColorBalance class.
// v0.22 (Mikeytown2, 28th August 2008):
// - Bug with rounding in preserve_luminosity code, casts to int for now, avoiding pixels with color blowout.
// v0.23 (Mikeytown2, 28th August 2008):
// - More speed, doubles are now floats
// v0.24 (Mikeytown2, 28th August 2008):
// - scaled preserve_luminosity code by 64 (256*64=16384; 360*64=23040)
// - changed (float) to f; thanks for the tip IanB!
// v0.25 (Mikeytown2, 4th September 2008):
// - added in broken 16bit code
// - added an option to switch between gimp's and cinepaint's behavior.
// v0.26 (Mikeytown2, 5th September 2008):
// - fixed error with clone_gimp's default value - Thanks Gavino

/* -------------------------------------------------------------------------------------- */

#include "windows.h"
#include "avisynth.h"
#include "math.h"

//prototypes
void gimp_hsl_to_rgb_8int (int *,int *,int *);
int  gimp_rgb_to_l_8int (int, int, int);
void gimp_rgb_to_hsl_8int (int *, int *, int *);
int  gimp_hsl_value_8int (float, float, float);



//ColorBalance Class
class ColorBalance : public GenericVideoFilter 
{
	boolean preserve_luminosity;
	boolean color16bit;
	double cyan_red[3];
	double magenta_green[3];
	double yellow_blue[3];

	//8 bit
	BYTE r8_lookup[256];
	BYTE g8_lookup[256];
	BYTE b8_lookup[256];
	void Create8BitLookupTables();

	//16 bit
	BYTE r16_lookup[65536];
	BYTE g16_lookup[65536];
	BYTE b16_lookup[65536];
	void Create16BitLookupTables();
	
	/*  for lightening 8bit  */
	static double  highlights8_add[256];
	static double  midtones8_add[256];
	static double  shadows8_add[256];
	/*  for darkening 8bit  */
	static double  highlights8_sub[256];
	static double  midtones8_sub[256];
	static double  shadows8_sub[256];


	/*  for lightening 16bit  */
	static double  highlights16_add[65536];
	static double  midtones16_add[65536];
	static double  shadows16_add[65536];
	/*  for darkening 16bit  */
	static double  highlights16_sub[65536];
	static double  midtones16_sub[65536];
	static double  shadows16_sub[65536];


	static void Transfer8BitInit(void);
	static void Transfer16BitInit(void);
	static boolean transfer8Bit_initialized;
	static boolean transfer16Bit_initialized;
	static bool like_gimp;
	enum mode {SHADOWS=0, MIDTONES=1, HIGHLIGHTS=2};
	static double SQR(double d) { return d*d; }
	static int CLAMP0255(int i) { return max(min(i, 255), 0); }
	static int CLAMP065535(int i) { return max(min(i, 65535), 0); }

public:
	//Public Interface
	ColorBalance(PClip _child,
				double rs, double gs, double bs, // shadows
				double rm, double gm, double bm, // midtones
				double rh, double gh, double bh, // highlights
				bool keep_luma, bool clone_gimp, bool highcolor,
				IScriptEnvironment* env);

	//Destructor
	~ColorBalance() {}
	virtual PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
	static AVSValue __cdecl Create(AVSValue args, void* user_data, IScriptEnvironment* env);
};

/*  for lightening 8bit  */
double ColorBalance::highlights8_add[256] = { 0 };
double ColorBalance::midtones8_add[256]   = { 0 };
double ColorBalance::shadows8_add[256]    = { 0 };

/*  for darkening 8bit  */
double ColorBalance::highlights8_sub[256] = { 0 };
double ColorBalance::midtones8_sub[256]   = { 0 };
double ColorBalance::shadows8_sub[256]    = { 0 };



/*  for lightening 16bit  */
double ColorBalance::highlights16_add[65536] = { 0 };
double ColorBalance::midtones16_add[65536]   = { 0 };
double ColorBalance::shadows16_add[65536]    = { 0 };

/*  for darkening 16bit  */
double ColorBalance::highlights16_sub[65536] = { 0 };
double ColorBalance::midtones16_sub[65536]   = { 0 };
double ColorBalance::shadows16_sub[65536]    = { 0 };



//set table not built
boolean ColorBalance::transfer8Bit_initialized = FALSE;
boolean ColorBalance::transfer16Bit_initialized = FALSE;

bool ColorBalance::like_gimp = true;

AVSValue __cdecl ColorBalance::Create(AVSValue args, void* user_data, IScriptEnvironment* env) 
{
	return new ColorBalance(	args[0].AsClip(),
								args[1].AsFloat(0.0), args[2].AsFloat(0.0), args[3].AsFloat(0.0),
								args[4].AsFloat(0.0), args[5].AsFloat(0.0), args[6].AsFloat(0.0),
								args[7].AsFloat(0.0), args[8].AsFloat(0.0), args[9].AsFloat(0.0),
								args[10].AsBool(true), args[11].AsBool(true), args[12].AsBool(false),
								env);
}

ColorBalance::ColorBalance(	PClip _child,
							double rs, double gs, double bs, // shadows
							double rm, double gm, double bm, // midtones
							double rh, double gh, double bh, // highlights
							bool keep_luma, bool clone_gimp, bool highcolor,
							IScriptEnvironment* env): GenericVideoFilter(_child) 
{
	if (!vi.IsRGB()) 
	{
		env->ThrowError("ColorBalance: only RGB clips supported");
	}


	cyan_red[SHADOWS]		= rs;
	magenta_green[SHADOWS]	= gs;
	yellow_blue[SHADOWS]	= bs;

	cyan_red[MIDTONES]		= rm;
	magenta_green[MIDTONES]	= gm;
	yellow_blue[MIDTONES]	= bm;

	cyan_red[HIGHLIGHTS]		= rh;
	magenta_green[HIGHLIGHTS]	= gh;
	yellow_blue[HIGHLIGHTS]		= bh;

	preserve_luminosity = keep_luma;
	color16bit = highcolor;

	if (highcolor)
	{
		if (! transfer16Bit_initialized) 
		{
			Transfer16BitInit ();
			transfer16Bit_initialized = TRUE;
		}
		Create16BitLookupTables();
	}
	else
	{
		if (clone_gimp != like_gimp)
		{
			transfer8Bit_initialized = FALSE;
			like_gimp = clone_gimp;
		}
		if (! transfer8Bit_initialized) 
		{
			Transfer8BitInit ();
			transfer8Bit_initialized = TRUE;
		}
		Create8BitLookupTables();
	}
}

void ColorBalance::Create8BitLookupTables()
{
	double *cyan_red_transfer[3];
	double *magenta_green_transfer[3];
	double *yellow_blue_transfer[3];
	int     r_n, g_n, b_n;

	/*  Set the transfer arrays  (for speed)  */
	cyan_red_transfer[SHADOWS] =
		(cyan_red[SHADOWS] > 0) ? shadows8_add : shadows8_sub;
	cyan_red_transfer[MIDTONES] =
		(cyan_red[MIDTONES] > 0) ? midtones8_add : midtones8_sub;
	cyan_red_transfer[HIGHLIGHTS] =
		(cyan_red[HIGHLIGHTS] > 0) ? highlights8_add : highlights8_sub;

	magenta_green_transfer[SHADOWS] =
		(magenta_green[SHADOWS] > 0) ? shadows8_add : shadows8_sub;
	magenta_green_transfer[MIDTONES] =
		(magenta_green[MIDTONES] > 0) ? midtones8_add : midtones8_sub;
	magenta_green_transfer[HIGHLIGHTS] =
		(magenta_green[HIGHLIGHTS] > 0) ? highlights8_add : highlights8_sub;

	yellow_blue_transfer[SHADOWS] =
		(yellow_blue[SHADOWS] > 0) ? shadows8_add : shadows8_sub;
	yellow_blue_transfer[MIDTONES] =
		(yellow_blue[MIDTONES] > 0) ? midtones8_add : midtones8_sub;
	yellow_blue_transfer[HIGHLIGHTS] =
		(yellow_blue[HIGHLIGHTS] > 0) ? highlights8_add : highlights8_sub;

	for (int i = 0; i < 256; i++)
	{
		r_n = i;
		g_n = i;
		b_n = i;


		r_n += (int) ( cyan_red[SHADOWS] * cyan_red_transfer[SHADOWS][r_n] );
		r_n = CLAMP0255 (r_n);
		r_n += (int) ( cyan_red[MIDTONES] * cyan_red_transfer[MIDTONES][r_n] );
		r_n = CLAMP0255 (r_n);
		r_n += (int) ( cyan_red[HIGHLIGHTS] * cyan_red_transfer[HIGHLIGHTS][r_n] );
		r_n = CLAMP0255 (r_n);

		g_n += (int) ( magenta_green[SHADOWS] * magenta_green_transfer[SHADOWS][g_n] );
		g_n = CLAMP0255 (g_n);
		g_n += (int) ( magenta_green[MIDTONES] * magenta_green_transfer[MIDTONES][g_n] );
		g_n = CLAMP0255 (g_n);
		g_n += (int) ( magenta_green[HIGHLIGHTS] * magenta_green_transfer[HIGHLIGHTS][g_n] );
		g_n = CLAMP0255 (g_n);

		b_n += (int) ( yellow_blue[SHADOWS] * yellow_blue_transfer[SHADOWS][b_n] );
		b_n = CLAMP0255 (b_n);
		b_n += (int) ( yellow_blue[MIDTONES] * yellow_blue_transfer[MIDTONES][b_n] );
		b_n = CLAMP0255 (b_n);
		b_n += (int) ( yellow_blue[HIGHLIGHTS] * yellow_blue_transfer[HIGHLIGHTS][b_n] );
		b_n = CLAMP0255 (b_n);


		r8_lookup[i] = (BYTE) r_n;
		g8_lookup[i] = (BYTE) g_n;
		b8_lookup[i] = (BYTE) b_n;
	}
}


void ColorBalance::Create16BitLookupTables()
{
	double *cyan_red_transfer[3];
	double *magenta_green_transfer[3];
	double *yellow_blue_transfer[3];
	int     r_n, g_n, b_n;

	/*  Set the transfer arrays  (for speed)  */
	cyan_red_transfer[SHADOWS] =
		(cyan_red[SHADOWS] > 0) ? shadows16_add : shadows16_sub;
	cyan_red_transfer[MIDTONES] =
		(cyan_red[MIDTONES] > 0) ? midtones16_add : midtones16_sub;
	cyan_red_transfer[HIGHLIGHTS] =
		(cyan_red[HIGHLIGHTS] > 0) ? highlights16_add : highlights16_sub;

	magenta_green_transfer[SHADOWS] =
		(magenta_green[SHADOWS] > 0) ? shadows16_add : shadows16_sub;
	magenta_green_transfer[MIDTONES] =
		(magenta_green[MIDTONES] > 0) ? midtones16_add : midtones16_sub;
	magenta_green_transfer[HIGHLIGHTS] =
		(magenta_green[HIGHLIGHTS] > 0) ? highlights16_add : highlights16_sub;

	yellow_blue_transfer[SHADOWS] =
		(yellow_blue[SHADOWS] > 0) ? shadows16_add : shadows16_sub;
	yellow_blue_transfer[MIDTONES] =
		(yellow_blue[MIDTONES] > 0) ? midtones16_add : midtones16_sub;
	yellow_blue_transfer[HIGHLIGHTS] =
		(yellow_blue[HIGHLIGHTS] > 0) ? highlights16_add : highlights16_sub;

	for (int i = 0; i < 65536; i++)
	{
		r_n = i;
		g_n = i;
		b_n = i;

		r_n += (int) ( cyan_red[SHADOWS] * cyan_red_transfer[SHADOWS][r_n] );  // 65535/255=257
		r_n = CLAMP065535 (r_n);
		r_n += (int) ( cyan_red[MIDTONES] * cyan_red_transfer[MIDTONES][r_n] );
		r_n = CLAMP065535 (r_n);
		r_n += (int) ( cyan_red[HIGHLIGHTS] * cyan_red_transfer[HIGHLIGHTS][r_n] );
		r_n = CLAMP065535 (r_n);

		g_n += (int) ( magenta_green[SHADOWS] * magenta_green_transfer[SHADOWS][g_n] );
		g_n = CLAMP065535 (g_n);
		g_n += (int) ( magenta_green[MIDTONES] * magenta_green_transfer[MIDTONES][g_n] );
		g_n = CLAMP065535 (g_n);
		g_n += (int) ( magenta_green[HIGHLIGHTS] * magenta_green_transfer[HIGHLIGHTS][g_n] );
		g_n = CLAMP065535 (g_n);

		b_n += (int) ( yellow_blue[SHADOWS] * yellow_blue_transfer[SHADOWS][b_n] );
		b_n = CLAMP065535 (b_n);
		b_n += (int) ( yellow_blue[MIDTONES] * yellow_blue_transfer[MIDTONES][b_n] );
		b_n = CLAMP065535 (b_n);
		b_n += (int) ( yellow_blue[HIGHLIGHTS] * yellow_blue_transfer[HIGHLIGHTS][b_n] );
		b_n = CLAMP065535 (b_n);

		r16_lookup[i] = (BYTE) r_n;
		g16_lookup[i] = (BYTE) g_n;
		b16_lookup[i] = (BYTE) b_n;
	}
}


void ColorBalance::Transfer8BitInit(void) 
{
	for (int i = 0; i < 256; i++)
	{
		double low = (1.075 - 1 / ((double) i / 16.0 + 1));
		double mid = 2.0/3.0 * (1 - SQR (((double) i - 127.0) / 127.0));

		shadows8_sub[255 - i]		= low;
		shadows8_add[i]				= like_gimp ? low : mid;

		midtones8_sub[i]			= mid;
		midtones8_add[i]			= mid;
		
		highlights8_sub[i]			= like_gimp ? low: mid;
		if (like_gimp)
		{
			highlights8_add[255 - i]	= low; 
		}
		else
		{
			highlights8_add[i]			= low; 
		}
	}
}

void ColorBalance::Transfer16BitInit(void) 
{
	for (int i = 0; i < 65536; i++)
	{
		double x = ((double)i/65535.0);   //x = scale i, 0 to 1

		double low = 1.075 - 1.0 /(.0625 * x + 1.0);
		double mid = 2.0/3.0 * (1.0 - SQR ( 2.0 * x - 1.0));

		shadows16_sub[65535 - i]	= low;
		shadows16_add[i]			= mid;
		
		midtones16_sub[i]			= mid;
		midtones16_add[i]			= mid;
		
		highlights16_sub[i]			= mid;
		highlights16_add[i]			= low;
	}
}

PVideoFrame __stdcall ColorBalance::GetFrame(int n, IScriptEnvironment* env) 
{
	PVideoFrame src = child->GetFrame(n, env);
	PVideoFrame dst = env->NewVideoFrame(vi);
	const unsigned char* srcp = src->GetReadPtr();
	unsigned char* dstp = dst->GetWritePtr();
	const int dst_pitch = dst->GetPitch();
	const int src_pitch = src->GetPitch();
	const int src_width = src->GetRowSize();
	const int src_height = src->GetHeight();
	int w, h;
	int r, g, b;//, a; //a no longer needed
	int r_n, g_n, b_n;

	//all if's outside loops for speed; 2 conditions so 4 loops total
	if (vi.IsRGB()) 
	{
		if (vi.IsRGB24())  // RGB24, w=3
		{
			if (preserve_luminosity) //preserve_luminosity
			{
				for (h=0; h < src_height;h++) 
				{
					for (w = 0; w < src_width; w+=3) 
					{
						r = *(srcp + w+2);
						g = *(srcp + w+1);
						b = *(srcp + w+0);

						r_n = r8_lookup[r];
						g_n = g8_lookup[g];
						b_n = b8_lookup[b];
						
						//preserve_luminosity
						gimp_rgb_to_hsl_8int (&r_n, &g_n, &b_n);
						b_n = gimp_rgb_to_l_8int (r, g, b);
						gimp_hsl_to_rgb_8int (&r_n, &g_n, &b_n);

						*(dstp + w + 2)	= (char) r_n;
						*(dstp + w + 1)	= (char) g_n;
						*(dstp + w + 0)	= (char) b_n;

					}
					srcp = srcp + src_pitch;
					dstp = dstp + dst_pitch;
				}
			}
			else //NOT preserve_luminosity
			{
				if (color16bit) //16bit
				{
					for (h=0; h < src_height;h++) 
					{
						for (w = 0; w < src_width; w+=3) 
						{
							dstp[w+2] = r16_lookup[srcp[w+2]];
							dstp[w+1] = r16_lookup[srcp[w+1]];
							dstp[w+0] = r16_lookup[srcp[w+0]];
						}
						srcp = srcp + src_pitch;
						dstp = dstp + dst_pitch;
					}
				}
				else //8bit
				{
					for (h=0; h < src_height;h++) 
					{
						for (w = 0; w < src_width; w+=3) 
						{
							dstp[w+2] = r8_lookup[srcp[w+2]];
							dstp[w+1] = g8_lookup[srcp[w+1]];
							dstp[w+0] = b8_lookup[srcp[w+0]];
						}
						srcp = srcp + src_pitch;
						dstp = dstp + dst_pitch;
					}
				}

			}
		}
		else // RGB32, w=4
		{
			if (preserve_luminosity) //preserve_luminosity
			{
				for (h=0; h < src_height;h++) 
				{
					for (w = 0; w < src_width; w+=4) 
					{
						//a = *(srcp + w+3);
						r = *(srcp + w+2);
						g = *(srcp + w+1);
						b = *(srcp + w+0);

						r_n = r8_lookup[r];
						g_n = g8_lookup[g];
						b_n = b8_lookup[b];

						//preserve_luminosity
						gimp_rgb_to_hsl_8int (&r_n, &g_n, &b_n);
						b_n = gimp_rgb_to_l_8int (r, g, b);
						gimp_hsl_to_rgb_8int (&r_n, &g_n, &b_n);

						//*(dstp + w + 3)	= a;
						dstp[w+3] = srcp[w+3]; //alpha channel direct copy
						*(dstp + w + 2)	= (char) r_n;
						*(dstp + w + 1)	= (char) g_n;
						*(dstp + w + 0)	= (char) b_n;

					}
					srcp = srcp + src_pitch;
					dstp = dstp + dst_pitch;
				}
			}
			else //NOT preserve_luminosity
			{
				if (color16bit) //16bit
				{
					for (h=0; h < src_height;h++) 
					{
						for (w = 0; w < src_width; w+=3) 
						{
							dstp[w+3] =				srcp[w+3]; //alpha channel direct copy
							dstp[w+2] = r16_lookup[	srcp[w+2]];
							dstp[w+1] = r16_lookup[	srcp[w+1]];
							dstp[w+0] = r16_lookup[	srcp[w+0]];
						}
						srcp = srcp + src_pitch;
						dstp = dstp + dst_pitch;
					}
				}
				else //8bit
				{
					for (h=0; h < src_height;h++) 
					{
						for (w = 0; w < src_width; w+=4) 
						{
							dstp[w+3] =				srcp[w+3]; //alpha channel direct copy
							dstp[w+2] = r8_lookup[	srcp[w+2]];
							dstp[w+1] = g8_lookup[	srcp[w+1]];
							dstp[w+0] = b8_lookup[	srcp[w+0]];
						}
						srcp = srcp + src_pitch;
						dstp = dstp + dst_pitch;
					}
				}
			}
		}
	}

	return dst;
}


/**
	* gimp_rgb_to_hsl_int:
	* @red: Red channel, returns Hue channel
	* @green: Green channel, returns Lightness channel
	* @blue: Blue channel, returns Saturation channel
	*
	* The arguments are pointers to int representing channel values in the
	* RGB colorspace, and the values pointed to are all in the range [0, 255].
	*
	* The function changes the arguments to point to the corresponding HLS
	* value with the values pointed to in the following ranges:  H [0, 360*64],
	* L [0, 255*64], S [0, 255*64].
**/
void
gimp_rgb_to_hsl_8int (	int *red, 
						int *green, 
						int *blue)
{
	int    r, g, b;
	float  h, s, l;
	int    minimum, maximum;
	int    delta;

	r = *red;
	g = *green;
	b = *blue;

	if (r > g)
	{
		maximum = max (r, b);
		minimum = min (g, b);
	}
	else
	{
		maximum = max (g, b);
		minimum = min (r, b);
	}

	l = (maximum + minimum) / 2.0f;

	if (maximum == minimum)
	{
		s = 0.0f;
		h = 0.0f;
	}
	else
	{
		delta = (maximum - minimum);

		if (l < 128)
		{
			s = 255 * (float) delta / (float) (maximum + minimum);
		}
		else
		{
			s = 255 * (float) delta / (float) (511 - maximum - minimum);
		}

		if (r == maximum)
		{
			h = (g - b) / (float) delta;
		}
		else if (g == maximum)
		{
			h = 2 + (b - r) / (float) delta;
		}
		else
		{
			h = 4 + (r - g) / (float) delta;
		}

		h = h * 42.5f;

		if (h < 0)
		{
			h += 255;
		}
		else if (h > 255)
		{
			h -= 255;
		}
	}

	*red   = (int) (h*64.0f);
	*green = (int) (s*64.0f);
	*blue  = (int) (l*64.0f);
}


/**
	* gimp_rgb_to_l_int:
	* @red: Red channel
	* @green: Green channel
	* @blue: Blue channel
	*
	* Calculates the lightness value of an RGB triplet with the formula
	* L = (max(R, G, B) + min (R, G, B)) / 2
	* Range [0 - 256*64]
	*
	* Return value: Luminance vaue corresponding to the input RGB value
**/
int
gimp_rgb_to_l_8int (	int red,
					int green,
					int blue)
{
	int minimum, maximum;

	if (red > green)
	{
		maximum = max (red,   blue);
		minimum = min (green, blue);
	}
	else
	{
		maximum = max (green, blue);
		minimum = min (red,   blue);
	}

	return (int) ((maximum + minimum) / 2.0f * 64.0f);
}


/**
 * gimp_hsl_to_rgb_int:
 * @hue: Hue channel, returns Red channel
 * @saturation: Saturation channel, returns Green channel
 * @lightness: Lightness channel, returns Blue channel
 *
 * The arguments are pointers to int, with the values pointed to in the
 * following ranges:  H [0, 360*64], L [0, 255*64], S [0, 255*64].
 *
 * The function changes the arguments to point to the RGB value
 * corresponding, with the returned values all in the range [0, 255].
 **/
void
gimp_hsl_to_rgb_8int (	int *hue,
						int *saturation,
						int *lightness)
{
	float h, s, l;

	h = (float) *hue/64.0f;
	s = (float) *saturation/64.0f;
	l = (float) *lightness/64.0f;

	if (s == 0)
	{
		/*  achromatic case  */
		*hue        = (int) (l);
		*lightness  = (int) (l);
		*saturation = (int) (l);
	}
	else
	{
		float m1, m2;

		if (l < 128)
		{
			m2 = (l * (255 + s)) / 65025.0f;
		}
		else
		{
			m2 = (l + s - (l * s) / 255.0f) / 255.0f;
		}

		m1 = (l / 127.5f) - m2;

		/*  chromatic case  */
		*hue        = gimp_hsl_value_8int (m1, m2, h + 85);
		*saturation = gimp_hsl_value_8int (m1, m2, h);
		*lightness  = gimp_hsl_value_8int (m1, m2, h - 85);
	}
}

int
gimp_hsl_value_8int (	float n1,
						float n2,
						float hue)
{
	float value;

	if (hue > 255)
	{
		hue -= 255;
	}
	else if (hue < 0)
	{
		hue += 255;
	}

	if (hue < 42.5f)
	{
		value = n1 + (n2 - n1) * (hue / 42.5f);
	}
	else if (hue < 127.5f)
	{
		value = n2;
	}
	else if (hue < 170)
	{
		value = n1 + (n2 - n1) * ((170 - hue) / 42.5f);
	}
	else
	{
		value = n1;
	}

	return (int) (value * 255.0f);
}


//Add dll to AviSynth Scripting Environment
extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) 
{
	env->AddFunction(	"ColorBalance", "c[rs]f[gs]f[bs]f[rm]f[gm]f[bm]f[rh]f[gh]f[bh]f[keep_luma]b[clone_gimp]b[highcolor]b",
						ColorBalance::Create, 0);
	return "'ColorBalance' plugin - clone of Gimp tool";
}
